#include "TilemapComponent.h"
#include "../utils/Logging.h"
#include "../utils/FileDialog.h"
#include "../utils/utils.h"
#include <stb_image.h>
#include <cstdio>
#include <cstring>
#include <GLFW/glfw3.h>

static int g_selectedTileIndex = 0;


// Constructor: initialize with a default grid (optional)
TilemapComponent::TilemapComponent(Object *owner)
    : Component(owner)
{
    // For example, create an empty 10x10 tilemap
    SetGridSize(glm::ivec2(10, 10));
    // Optionally set a default tile size
    m_TileSize = glm::ivec2(32, 32);
}

// Set grid dimensions and resize tile array; initially all -1 (empty)
void TilemapComponent::SetGridSize(const glm::ivec2 &size)
{
    m_GridSize = size;
    m_Tiles.assign(size.x * size.y, -1);
}

// Set individual tile at (x, y)
void TilemapComponent::SetTile(int x, int y, int tileIndex)
{
    if (x < 0 || y < 0 || x >= m_GridSize.x || y >= m_GridSize.y)
        return;
    m_Tiles[y * m_GridSize.x + x] = tileIndex;
}

int TilemapComponent::GetTile(int x, int y) const
{
    if (x < 0 || y < 0 || x >= m_GridSize.x || y >= m_GridSize.y)
        return -1;
    return m_Tiles[y * m_GridSize.x + x];
}

void TilemapComponent::SetTileSize(const glm::ivec2 &size)
{
    m_TileSize = size;
}

// Set atlas (tileset) settings; also update dimensions
void TilemapComponent::SetAtlas(const std::string &path, int atlasTileWidth, int atlasTileHeight)
{
    m_AtlasPath = path;
    m_AtlasTileWidth = atlasTileWidth;
    m_AtlasTileHeight = atlasTileHeight;
    UpdateAtlasDimensions();
}

// Helper: load atlas image info using stbi_info to compute columns and rows
void TilemapComponent::UpdateAtlasDimensions()
{
    if (m_AtlasPath.empty())
        return;
    int texW, texH, comp;
    if (stbi_info(m_AtlasPath.c_str(), &texW, &texH, &comp))
    {
        m_AtlasCols = texW / m_AtlasTileWidth;
        m_AtlasRows = texH / m_AtlasTileHeight;
    }
    else
    {
        m_AtlasCols = m_AtlasRows = 0;
        Logger::LogError("Failed to get atlas info: %s", m_AtlasPath.c_str());
    }
}

// Serialization: store grid size, tile size, tile data, and atlas settings.
void TilemapComponent::Save(YAML::Emitter &out) const
{
    out << YAML::BeginMap;
    out << YAML::Key << "type" << YAML::Value << "TilemapComponent";
    out << YAML::Key << "gridSize" << YAML::Value << std::vector<int>{m_GridSize.x, m_GridSize.y};
    out << YAML::Key << "tileSize" << YAML::Value << std::vector<int>{m_TileSize.x, m_TileSize.y};
    out << YAML::Key << "tiles" << YAML::Value << m_Tiles;
    out << YAML::Key << "atlasPath" << YAML::Value << m_AtlasPath;
    out << YAML::Key << "atlasTileSize" << YAML::Value << std::vector<int>{m_AtlasTileWidth, m_AtlasTileHeight};
    out << YAML::EndMap;
}

// Deserialization: load saved data
void TilemapComponent::Load(const YAML::Node &node)
{
    if (node["gridSize"])
    {
        auto vec = node["gridSize"].as<std::vector<int>>();
        if (vec.size() == 2)
            m_GridSize = glm::ivec2(vec[0], vec[1]);
    }
    if (node["tileSize"])
    {
        auto vec = node["tileSize"].as<std::vector<int>>();
        if (vec.size() == 2)
            m_TileSize = glm::ivec2(vec[0], vec[1]);
    }
    if (node["tiles"])
        m_Tiles = node["tiles"].as<std::vector<int>>();
    if (node["atlasPath"])
        m_AtlasPath = node["atlasPath"].as<std::string>();
    if (node["atlasTileSize"])
    {
        auto vec = node["atlasTileSize"].as<std::vector<int>>();
        if (vec.size() == 2)
        {
            m_AtlasTileWidth = vec[0];
            m_AtlasTileHeight = vec[1];
        }
    }
    // Update atlas dimensions after loading settings
    UpdateAtlasDimensions();
}

// -----------------------------
// Editor UI
// -----------------------------
void TilemapComponent::DrawEditorUI()
{
    ImGui::Begin("Tilemap Editor");

    // --- Atlas Settings ---
    if (ImGui::CollapsingHeader("Tileset Atlas Settings"))
    {
        // Display current atlas path
        ImGui::Text("Current Atlas: %s", m_AtlasPath.empty() ? "None" : m_AtlasPath.c_str());
        if (ImGui::Button("Import Atlas"))
        {
            std::string path = OpenFileDialog(FileDialogType::Images);
            if (!path.empty())
            {
                // For simplicity, default atlas tile dimensions are set via input later.
                m_AtlasPath = path;
                UpdateAtlasDimensions();
            }
        }
        // Configure atlas tile size
        int atlasW = m_AtlasTileWidth, atlasH = m_AtlasTileHeight;
        if (ImGui::InputInt("Atlas Tile Width", &atlasW))
        {
            m_AtlasTileWidth = atlasW;
            UpdateAtlasDimensions();
        }
        if (ImGui::InputInt("Atlas Tile Height", &atlasH))
        {
            m_AtlasTileHeight = atlasH;
            UpdateAtlasDimensions();
        }
        ImGui::Text("Atlas Grid: %d columns x %d rows", m_AtlasCols, m_AtlasRows);
    }

    // --- Map Grid Settings ---
    if (ImGui::CollapsingHeader("Map Settings"))
    {
        glm::ivec2 grid = m_GridSize;
        int gridW = grid.x, gridH = grid.y;
        if (ImGui::InputInt("Map Grid Width", &gridW) || ImGui::InputInt("Map Grid Height", &gridH))
        {
            SetGridSize(glm::ivec2(gridW, gridH));
        }
        glm::ivec2 ts = m_TileSize;
        int tileW = ts.x, tileH = ts.y;
        if (ImGui::InputInt("Tile Width", &tileW) || ImGui::InputInt("Tile Height", &tileH))
        {
            SetTileSize(glm::ivec2(tileW, tileH));
        }
    }

    // --- Tileset Palette Editor ---
    if (ImGui::CollapsingHeader("Tileset Palette"))
    {
        if (m_AtlasPath.empty())
        {
            ImGui::Text("No atlas loaded.");
        }
        else
        {
            // Retrieve texture dimensions via stbi_info (for preview UV calculation)
            int texW, texH, comp;
            if (stbi_info(m_AtlasPath.c_str(), &texW, &texH, &comp))
            {
                float uvTileW = float(m_AtlasTileWidth) / float(texW);
                float uvTileH = float(m_AtlasTileHeight) / float(texH);
                // Display each tile as an image button
                ImGui::Text("Select a tile:");
                for (int row = 0; row < m_AtlasRows; row++)
                {
                    for (int col = 0; col < m_AtlasCols; col++)
                    {
                        int tileIndex = row * m_AtlasCols + col;
                        // UV coordinates for this tile
                        float uvX = float(col) * uvTileW;
                        float uvY = float(row) * uvTileH;
                        ImVec2 uv0(uvX, uvY);
                        ImVec2 uv1(uvX + uvTileW, uvY + uvTileH);

                        char buf[32];
                        sprintf(buf, "##tile_%d", tileIndex);
                        // Display a button. Here, 32x32 pixels is used for preview size.
                        GLuint textureID = LoadTextureIfNeeded(m_AtlasPath);
                        if (ImGui::ImageButton(buf, (ImTextureID)(uintptr_t)textureID, ImVec2(32, 32), uv0, uv1))
                        {
                            g_selectedTileIndex = tileIndex;
                        }

                        ImGui::SameLine();
                    }
                    ImGui::NewLine();
                }
                ImGui::Text("Selected Tile: %d", g_selectedTileIndex);
            }
            else
            {
                ImGui::Text("Failed to load atlas info.");
            }
        }
    }

    // --- Map Editor ---
    if (ImGui::CollapsingHeader("Map Editor"))
    {
        if (m_GridSize.x == 0 || m_GridSize.y == 0)
        {
            ImGui::Text("No map created.");
        }
        else
        {
            ImGui::Text("Click a cell to paint with the selected tile.");
            ImGui::BeginChild("TilemapGrid", ImVec2(0, 300), true);
            for (int y = 0; y < m_GridSize.y; y++)
            {
                for (int x = 0; x < m_GridSize.x; x++)
                {
                    int tile = GetTile(x, y);
                    char label[32];
                    sprintf(label, "%d##%d_%d", tile, x, y);
                    if (ImGui::Button(label, ImVec2(30, 30)))
                    {
                        SetTile(x, y, g_selectedTileIndex);
                    }
                    ImGui::SameLine();
                }
                ImGui::NewLine();
            }
            ImGui::EndChild();
        }
    }

    ImGui::End();
}

