#include "Engine.h"
#include "Entitys/Object.h"
#include "Renderer.h"
#include "components/SpriteComponent.h"
#include "components/CameraComponent.h"
#include "components/LightComponent.h"
#include "components/TilemapComponent.h"


#include "utils/FileDialog.h"
#include "utils/Logging.h"

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <imgui.h>
#include <imgui_impl_glfw.h>
#include <imgui_impl_opengl3.h>

#include <functional>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <memory>
#include <filesystem>
#include <iostream>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <openssl/sha.h>

static std::vector<std::shared_ptr<Object>> objects;
static std::shared_ptr<Object> selected = nullptr;
static bool playing = false;

static glm::vec2 cameraPos = {0, 0};
static float cameraZoom = 1.0f;
static ImVec2 lastMousePos = {};

GLFWwindow *window = nullptr;

Engine::Engine()
{
    Init();
}

Engine::~Engine()
{
    Shutdown();
}

void Engine::ShowDebugOverlay(float deltaTime)
{
    static std::vector<float> fpsHistory;
    static const int maxHistory = 100;
    static float timer = 0;
    float fps = 1.0f / deltaTime;
    if (fpsHistory.size() >= maxHistory)
        fpsHistory.erase(fpsHistory.begin());
    if (timer >= 0.05)
    {
        timer = 0;
        fpsHistory.push_back(fps);
    }
    timer += deltaTime;

    ImGui::Begin("Performance Info", nullptr,
                 ImGuiWindowFlags_NoDecoration);

    ImGui::TextColored(ImVec4(0.3f, 1.0f, 0.3f, 1.0f), "Performance");
    ImGui::Separator();
    ImGui::Text("FPS:        %.1f", fps);
    ImGui::Text("Delta Time: %.4f s", deltaTime);
    ImGui::Text("Frame Time: %.2f ms", deltaTime * 1000.0f);
    ImGui::PlotLines("##FPS", fpsHistory.data(), fpsHistory.size(), 0, "FPS History", 0.0f, 144.0f, ImVec2(-1, 50));

    ImGui::Spacing();
    ImGui::TextColored(ImVec4(0.4f, 0.7f, 1.0f, 1.0f), "Renderer");
    ImGui::Separator();
    ImGui::Text("Draw Calls: %d", Renderer::GetDrawCallCount());
    ImGui::Text("Reserved Draws: %d", m_Reserved_draws);

    ImGui::Text("Camera Zoom: %.2f", cameraZoom);
    ImGui::Text("Camera Pos:  (%.1f, %.1f)", cameraPos.x, cameraPos.y);

    ImGui::End();
}

void Engine::Init()
{
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    window = glfwCreateWindow(1280, 720, "Create Editor", nullptr, nullptr);
    glfwMakeContextCurrent(window);
    glfwSwapInterval(0); // No VSync
    glewInit();

    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO &io = ImGui::GetIO();
    io.ConfigFlags |= ImGuiConfigFlags_DockingEnable;
    ImGui::StyleColorsDark();
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init("#version 330");
    m_Reserved_draws = 0;

    Renderer::Init();

    auto obj = std::make_shared<Object>("Hello, Create");

    objects.push_back(obj);
    selected = obj;
    Logger::LogInfo("Initialized Engine");
}
void DrawInspectorUI(std::shared_ptr<Object> selected)
{
    ImGui::Begin("Inspector");

    if (!selected)
    {
        ImGui::Text("No object selected.");
        ImGui::End();
        return;
    }

    // Static buffer and last-selected tracking
    static char nameBuffer[128] = "";
    static std::shared_ptr<Object> lastSelected = nullptr;

    // If a new object is selected, refresh buffer
    if (selected != lastSelected)
    {
        strncpy(nameBuffer, selected->GetName().c_str(), sizeof(nameBuffer));
        nameBuffer[sizeof(nameBuffer) - 1] = '\0'; // null terminate
        lastSelected = selected;
    }

    if (ImGui::InputText("Name", nameBuffer, sizeof(nameBuffer)))
        selected->SetName(nameBuffer);

    glm::vec2 pos = selected->GetLocalPosition();
    if (ImGui::DragFloat2("Position", &pos.x))
        selected->SetLocalPosition(pos);

    ImGui::InputInt("Layer", &selected->layer);

    ImGui::SeparatorText("Add Components");

    if (ImGui::Button("Add SpriteComponent"))
    {
        if (!selected->GetComponent<SpriteComponent>())
            selected->AddComponent<SpriteComponent>();
    }
    if (ImGui::Button("Add CameraComponent"))
    {
        if (!selected->GetComponent<CameraComponent>())
            selected->AddComponent<CameraComponent>();
    }
    if (ImGui::Button("Add LightComponent"))
    {
        if (!selected->GetComponent<LightComponent>())
            selected->AddComponent<LightComponent>();
    }
    if (ImGui::Button("Add TilemapComponent"))
    {
        if (!selected->GetComponent<TilemapComponent>())
            selected->AddComponent<TilemapComponent>();
    }

    // Sprite UI...
    if (auto sprite = selected->GetComponent<SpriteComponent>())
    {
        ImGui::SeparatorText("Sprite Component");
        std::string tex = sprite->GetTexturePath();
        std::string norm = sprite->GetNormalMapPath();

        ImGui::Text("Texture: %s", tex.c_str());
        if (ImGui::Button("Load Texture"))
        {
            auto path = OpenFileDialog(FileDialogType::Images);
            if (!path.empty())
                sprite->SetTexture(path);
        }

        ImGui::Text("Normal Map: %s", norm.c_str());
        if (ImGui::Button("Load Normal Map"))
        {
            auto path = OpenFileDialog(FileDialogType::Images);
            if (!path.empty())
                sprite->SetNormalMap(path);
        }

        if (ImGui::Button("Remove SpriteComponent"))
            selected->RemoveComponent<SpriteComponent>();
    }

    if (auto cam = selected->GetComponent<CameraComponent>())
    {
        ImGui::SeparatorText("Camera Component");

        float fov = cam->GetFOV();
        float aspect = cam->GetAspect();
        float zoom = cam->GetZoom();

        if (ImGui::DragFloat("FOV", &fov, 0.1f, 1.0f, 179.0f)) cam->SetFOV(fov);
        if (ImGui::DragFloat("Aspect", &aspect, 0.01f, 0.1f, 5.0f)) cam->SetAspect(aspect);
        if (ImGui::DragFloat("Zoom", &zoom, 0.1f, 0.1f, 10.0f)) cam->SetZoom(zoom);

        if (ImGui::Button("Remove CameraComponent"))
            selected->RemoveComponent<CameraComponent>();
    }

    // Light UI...
    if (auto light = selected->GetComponent<LightComponent>())
    {
        ImGui::SeparatorText("Light Component");

        glm::vec3 color = light->GetColor();
        float intensity = light->GetIntensity();
        float radius = light->GetRadius();
        float falloff = light->GetFalloff();
        int type = light->GetType();

        if (ImGui::ColorEdit3("Color", &color.x)) light->SetColor(color);
        if (ImGui::DragFloat("Intensity", &intensity, 0.05f, 0.0f, 10.0f)) light->SetIntensity(intensity);
        if (ImGui::DragFloat("Radius", &radius, 1.0f, 10.0f, 1000.0f)) light->SetRadius(radius);
        if (ImGui::DragFloat("Falloff", &falloff, 0.05f, 0.1f, 5.0f)) light->SetFalloff(falloff);

        //const char* types[] = { "Point", "Directional" };
        //if (ImGui::Combo("Type", &type, types, IM_ARRAYSIZE(types)))
        //    light->SetType(type);

        if (ImGui::Button("Remove LightComponent"))
            selected->RemoveComponent<LightComponent>();
    }

    if (auto tilemap = selected->GetComponent<TilemapComponent>())
    {
        ImGui::SeparatorText("Tilemap Component");

        ImGui::Text("Refer to Tilemap Editor Window");

        if (ImGui::Button("Remove TilemapComponent"))
            selected->RemoveComponent<TilemapComponent>();
    }

    ImGui::End();

    if (auto tilemap = selected->GetComponent<TilemapComponent>())
    {
        tilemap->DrawEditorUI();

    }
}


void Engine::Run()
{
    while (!glfwWindowShouldClose(window))
    {
        glfwPollEvents();

        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();
        ImGui::DockSpaceOverViewport(ImGui::GetMainViewport()->ID);

        float currentTime = glfwGetTime();
        static float lastTime = currentTime;
        float deltaTime = currentTime - lastTime;
        lastTime = currentTime;

        ShowDebugOverlay(deltaTime);
        Logger::Draw();

        if (ImGui::BeginMainMenuBar())
        {
            if (ImGui::Button(playing ? "Stop" : "Play"))
            {
                playing = !playing;
            }
            if (ImGui::BeginMenu("File"))
            {
                if (ImGui::MenuItem("Save Scene"))
                {
                    std::string file = OpenFileDialog(FileDialogType::Scenes); // reuse your dialog
                    if (!file.empty())
                        SaveScene(file);
                }
                if (ImGui::MenuItem("Load Scene"))
                {
                    std::string file = OpenFileDialog(FileDialogType::Scenes);
                    if (!file.empty())
                        LoadScene(file);
                }
                ImGui::EndMenu();
            }

            ImGui::EndMainMenuBar();
        }

        ImGui::Begin("Scene Tree");

        if (ImGui::BeginPopupContextWindow("SceneTreeContext", ImGuiPopupFlags_MouseButtonRight))
        {
            if (ImGui::MenuItem("Create New Object"))
            {
                auto obj = std::make_shared<Object>("NewObject");
                obj->AddComponent<SpriteComponent>();
                objects.push_back(obj);
                selected = obj;
                ImGui::OpenPopup("RenameObject");
            }
            ImGui::Separator();
            if (ImGui::MenuItem("Create New Sprite Object"))
            {
                auto obj = std::make_shared<Object>("NewSprite");
                obj->AddComponent<SpriteComponent>();
                objects.push_back(obj);
                selected = obj;
                ImGui::OpenPopup("RenameObject");
            }
            ImGui::EndPopup();
        }

        for (auto &obj : objects)
            if (!obj->GetParent()) // Only draw root nodes
                DrawObjectNode(obj);

        ImGui::End();

        DrawInspectorUI(selected);

        // Viewport
        ImGui::Begin("Viewport", nullptr, ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoScrollWithMouse);
        ImVec2 size = ImGui::GetContentRegionAvail();

        if (!playing && ImGui::IsWindowHovered())
        {
            if (ImGui::IsMouseDragging(ImGuiMouseButton_Right))
            {
                ImVec2 mouseDelta = ImGui::GetMouseDragDelta(ImGuiMouseButton_Right);
                ImGui::ResetMouseDragDelta(ImGuiMouseButton_Right);
                cameraPos -= glm::vec2(mouseDelta.x, mouseDelta.y) / cameraZoom;
            }

            float wheel = ImGui::GetIO().MouseWheel;
            if (wheel != 0.0f)
            {
                float zoomFactor = (wheel > 0) ? 1.1f : 1.0f / 1.1f;

                // Get mouse position in screen space
                ImVec2 mousePos = ImGui::GetMousePos();
                ImVec2 windowPos = ImGui::GetWindowPos();
                ImVec2 contentRegionMin = ImGui::GetWindowContentRegionMin();
                glm::vec2 relativeMouse = glm::vec2(mousePos.x - windowPos.x - contentRegionMin.x,
                                                    mousePos.y - windowPos.y - contentRegionMin.y);

                // Convert to world position before zoom
                glm::vec2 beforeZoom = (relativeMouse - glm::vec2(size.x / 2, size.y / 2)) / cameraZoom + cameraPos;

                // Apply zoom
                cameraZoom *= zoomFactor;
                cameraZoom = std::clamp(cameraZoom, 0.1f, 10.0f);

                // Convert to world position after zoom
                glm::vec2 afterZoom = (relativeMouse - glm::vec2(size.x / 2, size.y / 2)) / cameraZoom + cameraPos;

                // Offset camera so zoom is centered at mouse
                cameraPos += beforeZoom - afterZoom;
            }
        }

        Renderer::Resize((int)size.x, (int)size.y);
        Renderer::Begin();

        std::vector<std::shared_ptr<Object>> toDraw;

        toDraw.reserve(m_Reserved_draws);
        m_Reserved_draws = 0; // Reset

        Renderer::DrawEditorGrid(cameraPos, cameraZoom);

        std::function<void(const std::shared_ptr<Object> &)> collect = [&](const std::shared_ptr<Object> &obj)
        {
            toDraw.push_back(obj);
            m_Reserved_draws += 1;

            if (auto light = obj->GetComponent<LightComponent>())
            {
                glm::vec2 world = obj->GetWorldPosition();
                glm::vec2 screen = (world - cameraPos) * cameraZoom + glm::vec2(Renderer::GetSize().x * 0.5f, Renderer::GetSize().y * 0.5f);

                Renderer::AddLight(screen, light->GetColor(), light->GetIntensity(), light->GetRadius() * cameraZoom);
            }

            for (const auto &child : obj->GetChildren())
                collect(child);
        };

        for (const auto &obj : objects)
            if (!obj->GetParent())
                collect(obj);

        std::sort(toDraw.begin(), toDraw.end(), [](const auto &a, const auto &b)
                  {
                    if (a->layer != b->layer)
                        return a->layer < b->layer;
                
                    return a->GetWorldPosition().y < b->GetWorldPosition().y; });

        for (const auto &obj : toDraw)
        {
            if (auto sprite = obj->GetComponent<SpriteComponent>())
            {
                glm::vec2 worldPos = obj->GetWorldPosition();
                Renderer::DrawSprite(sprite.get(), worldPos, cameraZoom, cameraPos);
            }
        }

        Renderer::End();
        GLuint texID = Renderer::GetRenderTexture();
        ImGui::Image((ImTextureID)(uintptr_t)texID, size, ImVec2(0, 1), ImVec2(1, 0));
        ImGui::End();

        if (ImGui::BeginPopup("RenameObject"))
        {
            static char nameBuffer[128];
            static bool once = true;

            if (once && selected)
            {
                strcpy(nameBuffer, selected->GetName().c_str());
                once = false;
            }

            ImGui::InputText("##rename", nameBuffer, sizeof(nameBuffer));
            if (ImGui::IsKeyPressed(ImGuiKey_Enter) || ImGui::Button("OK"))
            {
                if (selected)
                    selected->SetName(nameBuffer);
                ImGui::CloseCurrentPopup();
                once = true;
            }

            ImGui::SameLine();
            if (ImGui::Button("Cancel"))
            {
                ImGui::CloseCurrentPopup();
                once = true;
            }

            ImGui::EndPopup();
        }

        // ImGui render
        ImGui::Render();
        int w, h;
        glfwGetFramebufferSize(window, &w, &h);
        glViewport(0, 0, w, h);
        glClear(GL_COLOR_BUFFER_BIT);
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());
        glfwSwapBuffers(window);
    }
}

void Engine::DrawObjectNode(const std::shared_ptr<Object> &obj)
{
    ImGuiTreeNodeFlags flags = ImGuiTreeNodeFlags_OpenOnArrow |
                               ImGuiTreeNodeFlags_SpanAvailWidth |
                               (obj == selected ? ImGuiTreeNodeFlags_Selected : 0);

    bool open = ImGui::TreeNodeEx((void *)(intptr_t)obj->id, flags, "%s", obj->GetName().c_str());

    if (ImGui::IsItemClicked())
        selected = obj;

    // === Context Menu on Object ===
    if (ImGui::BeginPopupContextItem())
    {
        if (ImGui::MenuItem("Rename"))
        {
            selected = obj;
            ImGui::OpenPopup("RenameObject");
        }
        if (ImGui::MenuItem("Delete"))
        {
            // Remove from parent or root
            if (obj->GetParent())
            {
                obj->GetParent()->RemoveChild(obj.get());
            }
            else
            {
                objects.erase(std::remove_if(objects.begin(), objects.end(),
                                             [&](const std::shared_ptr<Object> &o)
                                             { return o == obj; }),
                              objects.end());
            }
            ImGui::EndPopup();
            return; // Don't draw deleted node
        }
        if (ImGui::MenuItem("Create Child"))
        {
            auto child = std::make_shared<Object>("NewObject");
            obj->AddChild(child);
            selected = child;
            child->SetName("NewObject");
            ImGui::OpenPopup("RenameObject");
        }
        ImGui::EndPopup();
    }

    // === Drag and Drop ===
    if (ImGui::BeginDragDropSource())
    {
        ImGui::SetDragDropPayload("OBJECT", &obj, sizeof(obj));
        ImGui::Text("Move: %s", obj->GetName().c_str());
        ImGui::EndDragDropSource();
    }

    if (ImGui::BeginDragDropTarget())
    {
        if (const ImGuiPayload *payload = ImGui::AcceptDragDropPayload("OBJECT"))
        {
            auto dragged = *(std::shared_ptr<Object> *)payload->Data;
            if (dragged != obj)
                obj->AddChild(dragged);
        }
        ImGui::EndDragDropTarget();
    }

    // === Children ===
    if (open)
    {
        for (auto &child : obj->GetChildren())
            DrawObjectNode(child);
        ImGui::TreePop();
    }
}

void Engine::SaveScene(const std::string &path)
{
    YAML::Emitter out;
    YAML::Emitter sceneData;

    sceneData << YAML::BeginSeq;
    for (const auto &obj : objects)
        obj->Save(sceneData);
    sceneData << YAML::EndSeq;

    std::string sceneString = sceneData.c_str();
    unsigned char hash[SHA256_DIGEST_LENGTH];
    SHA256(reinterpret_cast<const unsigned char *>(sceneString.c_str()), sceneString.size(), hash);

    std::ostringstream hashHex;
    for (int i = 0; i < SHA256_DIGEST_LENGTH; ++i)
        hashHex << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(hash[i]);

    // Output full file
    out << YAML::BeginMap;
    out << YAML::Key << "engine_version" << YAML::Value << "0.1.0";
    out << YAML::Key << "scene_name" << YAML::Value << std::filesystem::path(path).stem().string();
    out << YAML::Key << "scene_hash" << YAML::Value << hashHex.str();
    out << YAML::Key << "format_version" << YAML::Value << 1;
    out << YAML::Key << "objects" << YAML::Value << YAML::Load(sceneString);
    out << YAML::EndMap;

    std::ofstream file(path);
    file << out.c_str();
}

bool VerifySceneHash(const YAML::Node &root)
{
    if (!root["scene_hash"] || !root["objects"])
        return false;

    YAML::Emitter sceneOnly;
    sceneOnly << root["objects"];

    std::string sceneString = sceneOnly.c_str();
    unsigned char hash[SHA256_DIGEST_LENGTH];
    SHA256(reinterpret_cast<const unsigned char *>(sceneString.c_str()), sceneString.size(), hash);

    std::ostringstream hashHex;
    for (int i = 0; i < SHA256_DIGEST_LENGTH; ++i)
        hashHex << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(hash[i]);

    return hashHex.str() == root["scene_hash"].as<std::string>();
}

void Engine::LoadScene(const std::string &path)
{
    YAML::Node root = YAML::LoadFile(path);

    if (!root["engine_version"] || !root["format_version"] || !root["scene_name"])
    {
        Logger::LogError("[LoadScene] Missing required metadata!");
        return;
    }

    if (root["engine_version"].as<std::string>() != "0.1.0")
    {
        Logger::LogError("[LoadScene] Version mismatch! Expected 0.1.0, got %s", root["engine_version"].as<std::string>().c_str());
        return;
    }

    if (!VerifySceneHash(root))
    {
        Logger::LogWarning("[LoadScene] Scene hash does not match! File may be corrupted or tampered.");
    }

    objects.clear();
    const auto &objectArray = root["objects"];
    for (const auto &node : objectArray)
    {
        auto obj = std::make_shared<Object>("[DefaultObject]");
        obj->Load(node);
        objects.push_back(obj);
    }

    Logger::LogInfo("[LoadScene] Loaded scene: %s", root["scene_name"].as<std::string>().c_str());
}

void Engine::Shutdown()
{
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();
    glfwDestroyWindow(window);
    glfwTerminate();
}
